import pkg_resources
from werkzeug.datastructures import FileStorage
from werkzeug.exceptions import MethodNotAllowed, NotFound
from flaschengeist.models.user import _Avatar, User
from flaschengeist.utils.hook import HookBefore, HookAfter

plugins_installed = HookAfter("plugins.installed")
"""Hook decorator for when all plugins are installed
    Possible use case would be to populate the database with some presets.

    Args:
        hook_result: void (kwargs)
"""
plugins_loaded = HookAfter("plugins.loaded")
"""Hook decorator for when all plugins are loaded
    Possible use case would be to check if a specific other plugin is loaded and change own behavior

    Args:
        app: Current flask app instance (args)
        hook_result: void (kwargs)
"""
before_role_updated = HookBefore("update_role")
"""Hook decorator for when roles are modified
Args:
    role: Role object to modify
    new_name: New name if the name was changed (None if delete)
"""
after_role_updated = HookAfter("update_role")
"""Hook decorator for when roles are modified
Args:
    role: Role object containing the modified role
    new_name: New name if the name was changed (None if deleted)
"""
before_update_user = HookBefore("update_user")
"""Hook decorator, when ever an user update is done, this is called before.
Args:
    user: User object
"""
before_delete_user = HookBefore("delete_user")
"""Hook decorator,this is called before an user gets deleted.
Args:
    user: User object
"""


class Plugin:
    """Base class for all Plugins
    If your class uses custom models add a static property called ``models``"""

    blueprint = None  # You have to override
    """Override with a `flask.blueprint` if the plugin uses custom routes"""
    permissions = []  # You have to override
    """Override to add custom permissions used by the plugin
    
    A good style is to name the permissions with a prefix related to the plugin name,
    to prevent clashes with other plugins. E. g. instead of *delete* use *plugin_delete*.
    """
    id = "dev.flaschengeist.plugin"  # You have to override
    """Override with the unique ID of the plugin (Hint: FQN)"""
    name = "plugin"  # You have to override
    """Override with human readable name of the plugin"""
    models = None  # You have to override
    """Override with models module"""
    migrations_path = None  # Override this with the location of your db migrations directory
    """Override with path to migration files, if custome db tables are used"""

    def __init__(self, config=None):
        """Constructor called by create_app
        Args:
            config: Dict configuration containing the plugin section
        """
        self.version = pkg_resources.get_distribution(self.__module__.split(".")[0]).version

    def install(self):
        """Installation routine

        Is always called with Flask application context
        """
        pass

    def get_setting(self, name: str, **kwargs):
        """Get plugin setting from database

        Args:
            name: string identifying the setting
            default: Default value
        Returns:
            Value stored in database (native python)
        Raises:
            `KeyError` if no such setting exists in the database
        """
        from ..controller import pluginController

        return pluginController.get_setting(self.id)

    def set_setting(self, name: str, value):
        """Save setting in database

        Args:
            name: String identifying the setting
            value: Value to be stored
        """
        from ..controller import pluginController

        return pluginController.set_setting(self.id, name, value)

    def notify(self, user, text: str, data=None):
        """Create a new notification for an user

        Args:
            user: `flaschengeist.models.user.User` to notify
            text: Visibile notification text
            data: Optional data passed to the notificaton
        Returns:
            ID of the created `flaschengeist.models.notification.Notification`

        Hint: use the data for frontend actions.
        """
        from ..controller import pluginController

        return pluginController.notify(self.id, user, text, data)

    def serialize(self):
        """Serialize a plugin into a dict

        Returns:
            Dict containing version and permissions of the plugin
        """
        return {"version": self.version, "permissions": self.permissions}


class AuthPlugin(Plugin):
    def login(self, user, pw):
        """Login routine, MUST BE IMPLEMENTED!

        Args:
            user: User class containing at least the uid
            pw: given password
        Returns:
            Must return False if not found or invalid credentials, True if success
        """
        raise NotImplemented

    def update_user(self, user):
        """If backend is using external data, then update this user instance with external data
        Args:
            user: User object
        """
        pass

    def find_user(self, userid, mail=None):
        """Find an user by userid or mail
        Args:
            userid: Userid to search
            mail: If set, mail to search
        Returns:
            None or User
        """
        return None

    def modify_user(self, user, password, new_password=None):
        """If backend is using (writeable) external data, then update the external database with the user provided.
        User might have roles not existing on the external database, so you might have to create those.

        Args:
            user: User object
            password: Password (some backends need the current password for changes) if None force edit (admin)
            new_password: If set a password change is requested
        Raises:
            NotImplemented: If backend does not support this feature (or no password change)
            BadRequest: Logic error, e.g. password is wrong.
            Error: Other errors if backend went mad (are not handled and will result in a 500 error)
        """
        raise NotImplemented

    def create_user(self, user, password):
        """If backend is using (writeable) external data, then create a new user on the external database.

        Args:
            user: User object
            password: string

        """
        raise MethodNotAllowed

    def delete_user(self, user):
        """If backend is using (writeable) external data, then delete the user from external database.

        Args:
            user: User object

        """
        raise MethodNotAllowed

    def get_avatar(self, user: User) -> _Avatar:
        """Retrieve avatar for given user (if supported by auth backend)

        Default behavior is to use native Image objects,
        so by default this function is never called, as the userController checks
        native Image objects first.

        Args:
            user: User to retrieve the avatar for
        Raises:
            NotFound: If no avatar found or not implemented
        """
        raise NotFound

    def set_avatar(self, user: User, file: FileStorage):
        """Set the avatar for given user (if supported by auth backend)

        Default behavior is to use native Image objects stored on the Flaschengeist server

        Args:
            user: User to set the avatar for
            file: FileStorage object uploaded by the user
        Raises:
            MethodNotAllowed: If not supported by Backend
            Any valid HTTP exception
        """
        # By default save the image to the avatar,
        # deleting would happen by unsetting it
        from ..controller import imageController

        user.avatar_ = imageController.upload_image(file)

    def delete_avatar(self, user: User):
        """Delete the avatar for given user (if supported by auth backend)

        Default behavior is to use the imageController and native Image objects.

        Args:
            user: Uset to delete the avatar for

        Raises:
            MethodNotAllowed: If not supported by Backend
        """
        user.avatar_ = None
